from ..objects.component import *

# This component provides the ability authentice with Kerberos HDFS based on user config
class Kerberos(Component):
    activation = Activation.CONFIG

    krbConf = None
    keytab = None
    hdfsUser = None

    def __init__(self, krbConf=None, keytab=None, hdfsUser=None):
        self.krbConf = krbConf
        self.keytab = keytab
        self.hdfsUser = hdfsUser

    # Appends copy commands to the Dockerfile in order to add the needed Kerberos files to the correct location based on config
    def appendDockerfile(self):
        dockerfile = ""
        if (self.krbConf is not None) and (self.keytab is not None):
            dockerfile += "COPY {krb} /etc/krb5.conf\n".format(krb=self.krbConf)
            dockerfile += "COPY {keytab} /krb/auth.keytab\n".format(keytab=self.keytab)
        return dockerfile

    # If run inside of docker, execute the init.sh script with the username in order to initialize krb auth
    def prependCommand(self, job, native):
        return "/./krb/init.sh {user}".format(user=self.hdfsUser) if native == False else None



from ..objects.component import *
from ..systems.generators import yaml

# This component provides the ability to bump the version number in the skelebot.yaml file
class Bump(Component):
    activation = Activation.PROJECT
    commands = ["bump"]

    # Parser for the command that bumps the semantic version in the skelebot.yaml
    def addParsers(self, subparsers):
        helpMessage = "Bump the skelebot.yaml project version"
        parser = subparsers.add_parser("bump", help=helpMessage)
        parser.add_argument("version", help="Select the version number that should be bumped", choices=["major", "minor", "patch"])
        return subparsers

    # Read the current version and bump it based on the version provided (major, minor, or patch) and persist the yaml
    def execute(self, config, args):

        oldVersion = config.version
        mmp = oldVersion.split(".")
        mmp[0] = str(int(mmp[0]) + 1) if (args.version == "major") else mmp[0]
        mmp[1] = str(int(mmp[1]) + 1) if (args.version == "minor") else (mmp[1] if (args.version == "patch") else "0")
        mmp[2] = str(int(mmp[2]) + 1) if (args.version == "patch") else "0"
        version = ".".join(mmp)
        yaml.saveVersion(version)

        print("Bumped " + args.version + " version. v" + oldVersion + " -> v" + version)



from ...objects.config import Config
from ...objects.job import Job

import os

FILE_PATH = "{path}/README.md"

# Builds the README based on the config
def buildREADME(config):

    readme = """# {name}
![Version](https://img.shields.io/badge/Version-{version}-brightgreen.svg)
![Documentation](https://img.shields.io/badge/Documentation-UNLINKED-red.svg)

{description}

---

## Contact
Project Maintainer: {maintainer} ({contact})"""
    readme = readme.format(name=config.name, description=config.description, version=config.version, maintainer=config.maintainer, contact=config.contact)

    readmeFile = open(FILE_PATH.format(path=os.getcwd()), "w")
    readmeFile.write(readme)
    readmeFile.close()



from .skeleYaml import SkeleYaml
from ..common import LANGUAGE_IMAGE

# The base object for the configurations for a Skelebot project
class Config(SkeleYaml):
    name = None
    description = None
    version = None
    maintainer = None
    contact = None
    language = None
    baseImage = None
    primaryJob = None
    ephemeral = None
    dependencies = None
    ignores = None
    jobs = None
    ports = None
    components = None
    params = None

    # Initialize the object with required values and set the components list to an empty list to start
    def __init__(self, name=None, description=None, version=None, maintainer=None, contact=None, language=None, baseImage=None,
                 primaryJob=None, ephemeral=None, dependencies=[], ignores=[], jobs=[], ports=[], components=[], params=[]):
        self.name = name
        self.description = description
        self.version = version
        self.maintainer = maintainer
        self.contact = contact
        self.language = language
        self.baseImage = baseImage
        self.primaryJob = primaryJob
        self.ephemeral = ephemeral
        self.dependencies = dependencies
        self.ignores = ignores
        self.jobs = jobs
        self.ports = ports
        self.components = components
        self.params = params

    # Adds extra logic to handle the components coversion to dict since the YAML and Object structures don't exactly match
    def toDict(self):
        dctComp = {}
        for component in self.components:
            componentDct = component.toDict()
            if (componentDct != {}):
                dctComp[component.__class__.__name__.lower()] = componentDct

        dct = super().toDict()
        dct["components"] = dctComp
        return dct

    # Returns the proper base image based on the values for language and kerberos, or returns the user defined base image
    def getBaseImage(self):

        if (self.baseImage):
            image = self.baseImage
        else:
            language = self.language if self.language != None else "NA"
            image = LANGUAGE_IMAGE[language]

            variant = "base"
            for component in self.components:
                if (component.__class__.__name__.lower() == "kerberos"):
                    variant = "krb"

            image = image[variant]

        return image

    def getImageName(self):
        return self.name.lower().replace(" ", "-")



from ...objects.config import Config
from ...objects.job import Job
from ...objects.param import Param
from ...objects.component import Activation
from ...components.componentFactory import ComponentFactory
from ...components.plugin import Plugin

import yaml
import os
import copy

COMPONENTS_ATTRIBUTE = "components"
VERSION_ATTRIBUTE = "version"
FILE_PATH = "{path}/skelebot.yaml"
VERSION_PATH = "{path}/VERSION"
ENV_FILE_PATH = "{path}/skelebot-{env}.yaml"

def loadVersion():
    version = "0.0.0"
    with open(VERSION_PATH.format(path=os.getcwd()), 'r') as file:
        version = file.read().replace("\n", "")
    return version

def saveVersion(version):
    with open(VERSION_PATH.format(path=os.getcwd()), 'w') as file:
        file.write(version)

# Attempts to load the skelebot.yaml file into a Config object along with all of the activated componenets
def loadConfig(env=None):

    yamlData = readYaml(env)

    config = None
    if (yamlData is None):
        config = Config()
    else:
        values = {}
        values[VERSION_ATTRIBUTE] = loadVersion()
        for attr, value in yamlData.items():
            if (attr in vars(Config)) and (attr != COMPONENTS_ATTRIBUTE) and (attr != VERSION_ATTRIBUTE):
                if (attr == "jobs"):
                    values[attr] = Job.loadList(value)
                elif (attr == "params"):
                    values[attr] = Param.loadList(value)
                else:
                    values[attr] = value

        config = Config(**values)

    config.components = loadComponents(yamlData)

    return config

# Given a Config object, this function will generate the skelebot.yaml file with the values in the object
def saveConfig(config):
    saveVersion(config.version)
    config.version = None
    yml = yaml.dump(config.toDict(), default_flow_style=False)

    with open(FILE_PATH.format(path=os.getcwd()), "w") as file:
        file.write(yml)

# Reads the skelebot.yaml file (and env override if present) from the current path and loads it into a dict if present
def readYaml(env=None):
    yamlData = None
    cwd = os.getcwd()
    cfgFile = FILE_PATH.format(path=cwd)
    if os.path.isfile(cfgFile):
        with open(cfgFile, 'r') as stream:
            yamlData = yaml.load(stream)
            if (env is not None):
                envFile = ENV_FILE_PATH.format(path=cwd, env=env)
                if os.path.isfile(envFile):
                    with open(envFile, 'r') as stream:
                        overrideYaml = yaml.load(stream)
                        yamlData = override(yamlData, overrideYaml)

    return yamlData

# Override one dictionary with data from another dictionary
def override(orig, over):
    merged = copy.deepcopy(orig)
    for k, v2 in over.items():
        merged[k] = copy.deepcopy(v2)
    return merged

# Parses the components section of skelebot.yaml config to generate the complete list of components
# for the project based on the active component list and each components Activation attribute
def loadComponents(yamlData):

    components = []
    componentFactory = ComponentFactory()
    if (yamlData is None):
        components = componentFactory.buildComponents([Activation.EMPTY, Activation.ALWAYS])
    else:
        compNames = []
        if (COMPONENTS_ATTRIBUTE in yamlData):
            yamlConfig = yamlData[COMPONENTS_ATTRIBUTE]
            for compName in yamlConfig:
                component = componentFactory.buildComponent(compName, yamlConfig[compName])
                if (component is not None):
                    components.append(component)
                    compNames.append(component.__class__.__name__)

        components.extend(componentFactory.buildComponents([Activation.PROJECT, Activation.ALWAYS], ignores=compNames))

    return components



from ..objects.component import *
from ..systems.execution import docker
from ..systems.generators import dockerfile, dockerignore

# This component provides a way to spin up the container and exec into it for debugging or testing
class Dexec(Component):
    activation = Activation.PROJECT
    commands = ["exec"]

    # Parser for the command that exec into the docker container
    def addParsers(self, subparsers):
        helpMessage = "Exec into the running Docker container"
        parser = subparsers.add_parser("exec", help=helpMessage)
        parser.add_argument("-m", "--map", action="store_true", help="Volume map the working directory onto the container")
        return subparsers

    # Generate the Dockerfile and dockerignore and build the docker image and exec into the container
    def execute(self, config, args):

        mappings = []
        if (args.map):
            mappings.append(".")

        docker.build(config)
        docker.run(config, "/bin/bash", "it", [], mappings, "exec")



from unittest import TestCase
from unittest import mock

import skelebot as sb

class TestExecutor(TestCase):

    @mock.patch('os.path.expanduser')
    @mock.patch('os.path.exists')
    @mock.patch('os.getcwd')
    @mock.patch('skelebot.systems.scaffolding.scaffolder.ComponentFactory')
    @mock.patch('skelebot.systems.scaffolding.scaffolder.promptUser')
    def test_execute_scaffold_abort(self, mock_prompt, mock_cFactory, mock_getcwd, mock_exists, mock_expanduser):
        mock_expanduser.return_value = "test/plugins"
        mock_exists.return_value = False
        mock_prompt.side_effect = ["test", "test proj", "sean", "email", "Python", False]

        try:
            sb.systems.scaffolding.scaffolder.scaffold(False)
            self.fail("Exception Expected")
        except Exception as exc:
            self.assertEqual(str(exc), "Aborting Scaffolding Process")

            mock_prompt.assert_any_call("Enter a PROJECT NAME")
            mock_prompt.assert_any_call("Enter a PROJECT DESCRIPTION")
            mock_prompt.assert_any_call("Enter a MAINTAINER NAME")
            mock_prompt.assert_any_call("Enter a CONTACT EMAIL")
            mock_prompt.assert_any_call("Enter a LANGUAGE", options=["Python", "R"])
            mock_prompt.assert_any_call("Confirm Skelebot Setup", boolean=True)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.path.exists')
    @mock.patch('os.getcwd')
    @mock.patch('os.makedirs')
    @mock.patch('skelebot.systems.scaffolding.scaffolder.ComponentFactory')
    @mock.patch('skelebot.systems.scaffolding.scaffolder.yaml')
    @mock.patch('skelebot.systems.scaffolding.scaffolder.promptUser')
    def test_execute_scaffold_existing_init(self, mock_prompt, mock_yaml, mock_cFactory, mock_makedirs, mock_getcwd, mock_exists, mock_expanduser):
        mock_expanduser.return_value = "test/plugins"
        mock_exists.return_value = False
        mock_prompt.side_effect = ["test", "test proj", "sean", "email", "Python", True]

        sb.systems.scaffolding.scaffolder.scaffold(True)

        mock_prompt.assert_any_call("Enter a PROJECT NAME")
        mock_prompt.assert_any_call("Enter a PROJECT DESCRIPTION")
        mock_prompt.assert_any_call("Enter a MAINTAINER NAME")
        mock_prompt.assert_any_call("Enter a CONTACT EMAIL")
        mock_prompt.assert_any_call("Enter a LANGUAGE", options=["Python", "R"])
        mock_prompt.assert_any_call("Confirm Skelebot Setup", boolean=True)

        mock_yaml.saveConfig.assert_called_once()

    @mock.patch('os.path.expanduser')
    @mock.patch('os.path.exists')
    @mock.patch('os.getcwd')
    @mock.patch('os.makedirs')
    @mock.patch('skelebot.systems.scaffolding.scaffolder.ComponentFactory')
    @mock.patch('skelebot.systems.scaffolding.scaffolder.dockerfile')
    @mock.patch('skelebot.systems.scaffolding.scaffolder.dockerignore')
    @mock.patch('skelebot.systems.scaffolding.scaffolder.readme')
    @mock.patch('skelebot.systems.scaffolding.scaffolder.yaml')
    @mock.patch('skelebot.systems.scaffolding.scaffolder.promptUser')
    def test_execute_scaffold(self, mock_prompt, mock_yaml, mock_readme, mock_dignore, mock_dockerfile, mock_cFactory, mock_makedirs, mock_getcwd, mock_exists, mock_expanduser):
        mock_expanduser.return_value = "test/plugins"
        mock_prompt.side_effect = ["test", "test proj", "sean", "email", "Python", True]

        sb.systems.scaffolding.scaffolder.scaffold()

        mock_prompt.assert_any_call("Enter a PROJECT NAME")
        mock_prompt.assert_any_call("Enter a PROJECT DESCRIPTION")
        mock_prompt.assert_any_call("Enter a MAINTAINER NAME")
        mock_prompt.assert_any_call("Enter a CONTACT EMAIL")
        mock_prompt.assert_any_call("Enter a LANGUAGE", options=["Python", "R"])
        mock_prompt.assert_any_call("Confirm Skelebot Setup", boolean=True)

        mock_makedirs.assert_any_call("config/", exist_ok=True)
        mock_makedirs.assert_any_call("data/", exist_ok=True)
        mock_makedirs.assert_any_call("models/", exist_ok=True)
        mock_makedirs.assert_any_call("notebooks/", exist_ok=True)
        mock_makedirs.assert_any_call("output/", exist_ok=True)
        mock_makedirs.assert_any_call("queries/", exist_ok=True)
        mock_makedirs.assert_any_call("src/jobs/", exist_ok=True)

        mock_dockerfile.buildDockerfile.assert_called_once()
        mock_dignore.buildDockerignore.assert_called_once()
        mock_readme.buildREADME.assert_called_once()
        mock_yaml.saveConfig.assert_called_once()

if __name__ == '__main__':
    unittest.main()



from unittest import TestCase
from unittest import mock

import skelebot as sb
import os

class TestKerberos(TestCase):

    def test_appendDockerfile(self):
        kerberos = sb.components.kerberos.Kerberos(krbConf="krb/krb5.conf", keytab="krb/me.keytab")

        dockerfile = kerberos.appendDockerfile()
        expected = """COPY krb/krb5.conf /etc/krb5.conf
COPY krb/me.keytab /krb/auth.keytab
"""

        self.assertEqual(dockerfile, expected)

if __name__ == '__main__':
    unittest.main()



from .skeleYaml import SkeleYaml

class Arg(SkeleYaml):
    name = None
    choices = None
    help = None

    def __init__(self, name, choices=None, help=None):
        self.name = name
        self.choices = choices
        self.help = help

    @classmethod
    def load(cls, config):
        return cls(**config)

    @classmethod
    def loadList(cls, config):
        args = []
        for element in config:
            args.append(cls.load(element))

        return args



from ...objects.config import Config
from ...objects.job import Job

import os

FILE_PATH = "{path}/.dockerignore"

# Builds the dockerignore based on the config
def buildDockerignore(config):

    # Setup the basics of all dockerignores
    docker  = "# This dockerignore was generated by Skelebot\n"
    docker += "# Editing this file manually is not advised as all changes will be overwritten during Skelebot execution\n"

    # Add the base level ignores for the project defined in the yaml
    for ignore in config.ignores:
        docker += "{ignore}\n".format(ignore=ignore)

    # Pull in any additional dockerignore updates from the components
    for component in config.components:
        docker += component.appendDockerignore()

    dockerignore = open(FILE_PATH.format(path=os.getcwd()), "w")
    dockerignore.write(docker)
    dockerignore.close()



from ..objects.component import *
from ..objects.skeleYaml import SkeleYaml

import shutil
import artifactory
import os

ERROR_ALREADY_PUSHED = """This artifact version has already been pushed.
Please bump the version before pushing (skelebot bump) or force push (-f)."""

class Artifact(SkeleYaml):
    name = None
    file = None

    def __init__(self, name, file):
        self.name = name
        self.file = file

# This component provides the ability to spin up Jupyter in Docker for any project
class Artifactory(Component):
    activation = Activation.CONFIG
    commands = ["push", "pull"]

    artifacts = None
    url = None
    repo = None
    path = None

    # Load the yaml dict structure into the Artifactory class object and child object list
    @classmethod
    def load(cls, config):
        artifactDicts = config["artifacts"]
        artifacts = []
        for artifact in artifactDicts:
            newArtifact = Artifact(artifact["name"], artifact["file"])
            artifacts.append(newArtifact)
        return cls(artifacts, config["url"], config["repo"], config["path"])

    def __init__(self, artifacts=[], url=None, repo=None, path=None):
        self.artifacts = artifacts
        self.url = url
        self.repo = repo
        self.path = path

    # Parser for the push and pull commands to push and pull artifacts to/from Artifactory
    def addParsers(self, subparsers):
        artifactNames = []
        for artifact in self.artifacts:
            artifactNames.append(artifact.name)

        parser = subparsers.add_parser("push", help="Push an artifact to artifactory")
        parser.add_argument("artifact", choices=artifactNames)
        parser.add_argument("-u", "--user", help="Auth user for Artifactory")
        parser.add_argument("-t", "--token", help="Auth token for Artifactory")
        parser.add_argument("-f", "--force", action='store_true', help="Force the artifact to be pushed")

        parser = subparsers.add_parser("pull", help="Pull an artifact from artifactory")
        parser.add_argument("artifact", choices=artifactNames)
        parser.add_argument("version", help="The version of the artifact to pull")
        parser.add_argument("-u", "--user", help="Auth user for Artifactory")
        parser.add_argument("-t", "--token", help="Auth token for Artifactory")

        return subparsers

    # Generate the Dockerfile and dockerignore and build the docker image
    def execute(self, config, args):
        # Get User and Token if not provided in args
        user = None
        token = None
        if (args.user == None):
            user = input("Please provide a valid Artifactory user: ")

        if (args.token == None):
            token = input("Please provide a valid Artifactory token: ")

        # Obtain the artifact that matches the provided name
        selectedArtifact = None
        for artifact in self.artifacts:
            if (artifact.name == args.artifact):
                selectedArtifact = artifact

        # Generate the local artifact file and the final Artifactory url path
        ext = selectedArtifact.file.split(".")[1]
        version = config.version if (args.job == "push") else args.version
        file = "{filename}_v{version}.{ext}".format(filename=selectedArtifact.name, version=version, ext=ext)
        url = "{url}/{repo}/{path}/{file}".format(url=self.url, repo=self.repo, path=self.path, file=file)

        # Push the artifact with the config version, or pull with the arg version
        if (args.job == "push"):
            self.pushArtifact(selectedArtifact.file, user, token, file, url, args.force)
        elif (args.job == "pull"):
            self.pullArtifact(user, token, file, url)

    def pushArtifact(self, artifactFile, user, token, file, url, force):

        # Error and exit if artifact already exists and we are not forcing an override
        if (force == False) and artifactory.ArtifactoryPath(url).exists():
            raise Exception(ERROR_ALREADY_PUSHED)

        # Rename artifact, deploy the renamed artifact, and then rename it back to original name
        print("Deploying {file} to {url}".format(file=file, url=url))
        path = artifactory.ArtifactoryPath(url, auth=(user, token))
        shutil.copyfile(artifactFile, file)
        try:
            path.deploy_file(file)
        except:
            raise
        finally:
            os.remove(file)

    def pullArtifact(self, user, token, file, url):

        # Pull the artifact with the given url and save it to the file path
        if (artifactory.ArtifactoryPath(url).exists()):
            print("Pulling {file} from {url}".format(file=file, url=url))
            path = artifactory.ArtifactoryPath(url, auth=(user, token))
            with path.open() as fd:
                with open(file, "wb") as out:
                    content = fd.read()
                    print(content)
                    out.write(fd.read())
        else:
            print("Artifact Not Found: {url}".format(url=url))



from ...common import EXT_COMMAND

COMMAND = "{pre}{extCommand}{file}{params}{post}"

def build(config, job, args, native=False):

    ext = job.source.split(".")[1]
    args = vars(args)
    arguments = ""
    pre = ""
    post = ""

    # Build the params list from the args and params of the job with the supplied values (or defaults)
    params = buildArgs(job.args, args)
    params += buildParams(job.params, args)
    params += buildParams(config.params, args)

    # Construct the pre-run and post-run commands from the components
    for component in config.components:
        preCmd = component.prependCommand(job, native)
        postCmd = component.appendCommand(job, native)
        if (preCmd is not None):
            pre += "{cmd} && ".format(cmd=preCmd)
        if (postCmd is not None):
            post += " && {cmd}".format(cmd=postCmd)

    return COMMAND.format(extCommand=EXT_COMMAND[ext], file=job.source, args=arguments, params=params, pre=pre, post=post)

def buildArgs(argParams, args):
    argString = ""
    if (argParams != None):
        for arg in argParams:
            value = args.get(arg.name)
            if (value != None):
                argString += " {}".format(value)

    return argString

def buildParams(jobParams, args):
    paramString = ""
    if (jobParams != None):
        for param in jobParams:
            value = args.get(param.name, param.default)
            if (param.accepts == "boolean"):
                if (value == True):
                    paramString += " --{name}".format(name=param.name)
            elif (value != None):
                paramString += " --{name} {value}".format(name=param.name, value=value)

    return paramString
